/**
 * All hope abandon ye who enter here.
 * Refer to the documentation to maintain your sanity. - Dom
 *
 * All those that have come here have no sanity to maintain - Preston
 */

/**
 *	@namespace
 *	@desc Engine contains all functionality to create a game.
 */
var ENGINE = {};


/**
 * The total number of assets that can be used in the scene
 * @memberOf ENGINE
 * @type Number
 */
ENGINE.totalAssets = 0;
 
/**
 * The number of assets currently loaded in the scene
 * @memberOf ENGINE
 * @type Number
 */
ENGINE.assetsLoaded = 0;

/**
 * The current scene
 * @memberOf ENGINE
 * @type ENGINE.Scene
 */
ENGINE.currentScene = null;

ENGINE.maxTimeStep = 0.1;

/**
 * Static Input
 * @memberOf ENGINE
 * @readonly
 * @type Input
 */
ENGINE.Input = null;
// Create an entry point for window callbacks

ENGINE._elapsedTime = 0;
ENGINE._entryPoint = function()
{
	requestAnimationFrame(ENGINE._entryPoint);
	if(ENGINE.currentScene){
		var delta = Math.min(ENGINE.currentScene.clock.getDelta(), ENGINE.maxTimeStep);
		
		var elapsedTime = ENGINE._elapsedTime + delta; //ENGINE.currentScene.clock.elapsedTime; // BUG in THREE/ BAD DOCUMENTATION. time is in seconds, not ms

		//console.log('FPS: ' + (1/delta));

		ENGINE.GameTime = new GameTime(elapsedTime, delta);
		//THREE.AnimationHandler.update(delta); // Update animation
		
		ENGINE.currentScene.superEarlyUpdateRecursive(ENGINE.currentScene.scene);
		ENGINE.currentScene.earlyUpdateRecursive(ENGINE.currentScene.scene);
		ENGINE.currentScene.updateRecursive(ENGINE.currentScene.scene);
		ENGINE.currentScene.lateUpdateRecursive(ENGINE.currentScene.scene);
		
		// ***** Hack for call into 2D system ****
		if(game)
			if (game.main)
				if (game.main.updatePlayerInfo)
					game.main.updatePlayerInfo();
		// **********End Hack **********

		ENGINE.currentScene.draw();
	}
	ENGINE.Input.gamepadSupport.pollStatus();
	ENGINE.Input.updatePreviousKeyState(); // Update Input AFTER currentKeyState has changed since last frame
};

/**
    Starts the engine
	@param {Number} width The initial width of the canvas in pixels.
	@param {Number} height The initial height of the canvas in pixels.
	@param {Object} container The element to attach the canvas to.
	@param {Boolean} Should we use the deferred shader.
*/
ENGINE.start = function(width, height, container,deferredShading){
	// start the renderer and attach to container

	var scale = 1.0;
	ENGINE.renderer = new THREE.WebGLDeferredRenderer( { width: width, height: height, scale: scale, antialias: true, tonemapping: THREE.UnchartedOperator, brightness: 8 } );
	
	ENGINE.renderer.shadowMapEnabled = true;
	ENGINE.renderer.shadowMapType = THREE./*PCFSoft*/BasicShadowMap;
	
	container.appendChild(ENGINE.renderer.domElement);
	ENGINE.renderer.domElement.oncontextmenu=function(){return false;} // Disable context menu
	
	ENGINE.Input = new Input(ENGINE.renderer.domElement); // Initialize the input
	ENGINE._entryPoint(); // Start the main game loop
	
	//Window resize method, just resize the drawing area when the window resizes
	//window.onresize = function(){
	//	ENGINE.renderer.setSize(window.innerWidth,window.innerHeight);
	//}
};


// Namespace Object3D Here

/** @namespace THREE 
 *	@desc Additional functionality to the THREE namespace.
 */

/**
 *	@class
 *	@classdesc Additional functionality added to the THREE.Object3D class.
 *	@name Object3D
 *	@memberof THREE
 */

/**
 *	Adds a component to the GameObject.
 *	@param {Component} component The component to add.
 */
THREE.Object3D.prototype.addComponent = function(component){
	
	// CHECK FOR TYPE Property
	if(!component.type)
	{
		throw Error("Component Must Have Property 'type'");
		return;
	}
	
	if(this.components[component.type])
	{
		throw Error("Object already has component of type:" + component.type);
		return;
	}
	
	// Add Component
	this.components[component.type] = component;
	
	// Set Default Blank Mehtods
	if(!component.superEarlyUpdate)
		component.superEarlyUpdate = function(){};
	if(!component.earlyUpdate)
		component.earlyUpdate = function(){};
	if(!component.update)
		component.update = function(){};
	if(!component.lateUpdate)
		component.lateUpdate = function(){};
	if(!component.destroy)
		component.destroy = function(){};
	if(!component.init)
		component.init = function(){};
		
	component.gameObject = this;
};

/**
 *	Removes a component from the GameObject.
 *	@param {string} type The type of the component to be removed.
 */
THREE.Object3D.prototype.removeComponent = function(type){
	delete this.components[type];
};

/**
 *	Get a component attached to the GameObject.
 *	@param {string} type The type of the component to get.
 *	@returns {Component} The component retrieved.
 */
THREE.Object3D.prototype.getComponent = function(type){
	if(this.components[type] === undefined)
		return null;
	else
		return this.components[type];
};

// TODO GetComponentInChildren single component using depth first search? just return 1?

/**
 *	Gets all components of a specific type within the GameObject or any of its children.
 *	@param {string} type The type of the components to be retrieved.
 *	@returns {Array} The array of components found.
 */
THREE.Object3D.prototype.getComponentsInChildren = function(type){
	var components = [];
	var stack = [];
	stack.push(this);
	while (stack.length > 0)
	{
		// Remove current object from stack and add its component of type
		var currentObject = stack.pop();
		var currentComponent = currentObject.getComponent(type);
		if (currentComponent != null)
		{
			components.push(currentComponent);
		}
		stack = stack.concat(currentObject.children);
	}
	return components;
};

// TODO IMPORTANT: Override object add(object3d) remove(object3d) method to also INIT/delete COMPONENTS
var _old_add = THREE.Object3D.prototype.add;
THREE.Object3D.prototype.add = function(gameObject){
	
	this.old_add = _old_add;
	
	gameObject.castShadow = true; // EVERYTHING CASTS SHADOWS. HACK FOR LIGHTS

	this.old_add(gameObject);
	delete this.old_add;
	for(componentType in gameObject.components)
	{
		gameObject.components[componentType].init();
	}
};

var _old_remove = THREE.Object3D.prototype.remove;
/**
 *	Removes a child Object3D.
 *	@param {THREE.Object3D} gameObject The Object3D to remove.
 */
THREE.Object3D.prototype.remove = function(gameObject){
	for(componentType in gameObject.components)
	{
		gameObject.components[componentType].destroy();
	}
	this.old_remove = _old_remove;
	if(!(gameObject instanceof THREE.Light))
	{
		this.old_remove(gameObject);
	}
	delete this.old_remove;
};

// Add Components attribute tot he Object3D class
var _old_Object3D = THREE.Object3D;
THREE.Object3D = function()
{
	this.old_Object3D = _old_Object3D; // Doing this to get the right 'this' context
	this.old_Object3D();
	delete this.old_Object3D;			// All done, don't need this sticking around
	this.components = {};
	/** 
	 * 	Tag associated with the Object3D.
	 *	@type string
	 *	@memberof THREE.Object3D
	 *	@default ""
	 */
	this.tag = "";
};
THREE.Object3D.prototype = _old_Object3D.prototype;


/**
    GameTime Class
	@class
	@classdesc You will never have to make a new GameTime. Please refer to the instance at ENGINE.GameTime instead.
	@param {Number} timeSinceSceneLoad - The time in seconds since the level has loaded.
	@param {Number} deltaTime - The time in seconds it took to complete the last frame.
*/
var GameTime = function(timeSinceSceneLoad, deltaTime) {
	
	// Private
	var _timeSinceSceneLoad = timeSinceSceneLoad;
	var _deltaTime = deltaTime;

	//Public

	/** Get the time in seconds since the level has loaded.
	 *  @returns {Number}
	 */
	this.timeSinceSceneLoad = function(){return _timeSinceSceneLoad;};

	/** Get the time in seconds it took to complete the last frame. 
	 * @returns {Number}
	 */
	this.deltaTime = function(){return _deltaTime;};
};

/**
 *	GameTime
 *	@readonly
 *  @type GameTime
 */
ENGINE.GameTime = new GameTime(0,0);

/**
 *	Scene Class
 *	@class
 *	@classdesc A single game scene.  Be sure to set ENGINE.currentScene to your desired scene object.
 */
ENGINE.Scene = function(){
	
	// TODO: make Readonly clock
	this.clock = new THREE.Clock(true);
	
	/** The underlying THREE.Scene Object.
	 *	@readonly
	 */
	this.scene = new THREE.Scene(); // Same here

	/** The main camera used to draw to the canvas. */
	this.mainCamera = null;
	
	this.curLevel = 0;

	// TODO: make assets immutable. getters only
	/** All assets that can be used by objects in the scene 
	 * @readonly
	 */
	this.assets = {
		/** All geometry data */
		meshes: {},
		/** All texture data */
		textures: {},
		/** All audio data */
		sounds: {},
	};

	/** 
	 * Loads all assets to be used in the scene.
	 * @example
	 * // Assets Required Structure
	 * assetsRequired = [
	 *	 {
	 *		 "name": "AwesomeModel",
	 *		 "type":"mesh",
	 *		 "location": "http://www.awesomesite.com/game/AwesomeModel.fbx"
	 *	 },
	 *	 {
	 *		 "name": "AwesomeSound",
	 *		 "type":"sound",
	 *		 "location": "http://www.awesomesite.com/game/Sound.wav"
	 *	 },
	 *	 {
	 *		 "name": "AwesomeTexture",
	 *		 "type":"texture",
	 *		 "location": "http://www.awesomesite.com/game/texture.png"
	 *	 },
	 * ];
	 * @param {Array} assetsRequired Array of assets to be loaded
	 * @param {ENGINE.Scene} lastScene (Optional) Reference to the last scene
	 * @param {Function} loadCallback (Optional) Function to call after every asset is loaded
	 */
	this._currentAssetName = null; // Store the asset name for use in callback
	this.loadAssets = function(assetsRequired, lastScene, loadCallback){
	
		ENGINE.totalAssets = assetsRequired.length;
		
		// Loaders
		var jsonLoader = new THREE.JSONLoader();
		
		for (var i = 0; i < assetsRequired.length; i++){
		
			// Check to see if we already have this asset in memory. Copy a reference over if we do.
			if(lastScene !== undefined)
			{
				switch(assetsRequired[i].type)
				{
				case "mesh":
					if(lastScene.assets.meshes[assetsRequired[i].name])
					{
						ENGINE.currentScene.assets.meshes[assetsRequired[i].name] = lastScene.assets.meshes[assetsRequired[i].name];
						ENGINE.assetsLoaded = ENGINE.assetsLoaded + 1;
						if(loadCallback !== undefined)
						{
							loadCallback();
						}
						continue; // Move along assetsRequired[i], theres nothing to see here.
					}
					break;
				case "texture":
					if(lastScene.assets.textures[assetsRequired[i].name])
					{
						ENGINE.currentScene.assets.textures[assetsRequired[i].name] = lastScene.assets.textures[assetsRequired[i].name];
						ENGINE.assetsLoaded = ENGINE.assetsLoaded + 1;
						
						if(loadCallback !== undefined)
						{
							loadCallback();
						}
						continue; // Move along assetsRequired[i], theres nothing to see here.
					}
					break;
				case "sound":
					// TODO
					break;
				}
			}
			
			switch(assetsRequired[i].type)
			{
			// TODO: register callbacks to to load and error events to keep track of when finished
			case "mesh":
				// LOAD THE ASSET SYNCHROUNOUSLY!!! NOT ASYNC!
				var xhr = new XMLHttpRequest();
				xhr.open( "GET", assetsRequired[i].location, false );
				xhr.send();

				// Extract the texturepath if not explicitly given
				var texturePath = assetsRequired[i].texture_path ? 
						assetsRequired[i].texture_path : 
						jsonLoader.extractUrlBase( assetsRequired[i].location );
				
				// Store the asset name for use in callback
				this._currentAssetName = assetsRequired[i].name;
				
				// Callback function
				var callback = function(geometry, materials){
					// Store in assets
					ENGINE.currentScene.assets.meshes[ENGINE.currentScene._currentAssetName] = {
						'geometry': geometry, 
						'materials': materials,
					};
					
				//	console.log(ENGINE.currentScene._currentAssetName);
					
					// Set some defaults for each material
					for (var j = 0; j < materials.length; j++)
					{
						materials[j].skinning = true; // HAS TO BE FALSE WHEN NOT SKINNING
						materials[j].transparent = true;
						materials[j].alphaTest = 0.75;
					}
					
					// Reset currentAsset name
					ENGINE.currentScene._currentAssetName = null; 
					
					ENGINE.assetsLoaded = ENGINE.assetsLoaded + 1;
					if(loadCallback !== undefined)
					{
						loadCallback();
					}
				};
				// Create the model from the json
				
				
				var ext = assetsRequired[i].location.substr(assetsRequired[i].location.lastIndexOf('.') + 1);
				
				if(ext == "js")
				{
					var modelData = JSON.parse(xhr.responseText);
					jsonLoader.createModel( modelData, callback, texturePath );
				}
				
				break;
			case "texture":
				this.assets.textures[assetsRequired[i].name] = THREE.ImageUtils.loadTexture(assetsRequired[i].location);
				ENGINE.assetsLoaded = ENGINE.assetsLoaded + 1;
				if(loadCallback !== undefined)
				{
					loadCallback();
				}
				break;
			case "sound":
				//TODO: Load Sound
				break;
			}
		}
	};
	
	// TODO: make private somehow Private Update all given the scene
	
	this.updateRecursive = function(gameObject){
		for(component in gameObject.components) {
			gameObject.components[component].update(ENGINE.GameTime);
		}
		for(child in gameObject.children){
			this.updateRecursive(gameObject.children[child]);
		}
	};
	this.superEarlyUpdateRecursive = function(gameObject){
		for(component in gameObject.components) {
			gameObject.components[component].superEarlyUpdate(ENGINE.GameTime);
		}
		for(child in gameObject.children){
			this.superEarlyUpdateRecursive(gameObject.children[child]);
		}
	};
	this.earlyUpdateRecursive = function(gameObject){
		for(component in gameObject.components) {
			gameObject.components[component].earlyUpdate(ENGINE.GameTime);
		}
		for(child in gameObject.children){
			this.earlyUpdateRecursive(gameObject.children[child]);
		}
		
	
	};
	this.lateUpdateRecursive = function(gameObject){
		
		for(component in gameObject.components) {
			gameObject.components[component].lateUpdate(ENGINE.GameTime);
		}
		for(child in gameObject.children){
			this.lateUpdateRecursive(gameObject.children[child]);
		}
		
		if(Audio.ready)
		{
			Audio.Update(ENGINE.currentScene.mainCamera.matrix);
		}
	};
	
	
	
	// Private
	this.draw = function(){
		// TODO: Maybe resize to window here?
		if (this.mainCamera)
		{
			ENGINE.renderer.render(this.scene, this.mainCamera);
		}
	};

	/** 
	 *	Adds an Object3D to the root level of the scene.
	 *	@param {THREE.Object3D} gameObject The object to be added.
	 */
	this.add = function(gameObject){
		this.scene.add(gameObject);
	};
	
	/** 
	 *	Removes an Object3D from the scene.
	 *	@param {THREE.Object3D} gameObject The object to be removed.
	 */
	this.remove = function(gameObject){
		gameObject.parent.remove(gameObject);
	};
	
	/** 
	 *	Removes an Object3D from the scene by name.
	 *	@param {string} name The name of the object to be removed.
	 */
	this.removeByName = function(name){
		var gameObject = this.scene.getChildByName(name, true);
		this.remove(gameObject);
	};
	
	/** 
	 *	Find an Object3D in the scene by name
	 *	@param {string} name The name of the object to be retrieved.
	 *  @returns {Object3D} The object found.
	 */
	this.find = function(name){
		return this.scene.getChildByName(name, true);
	};
	
	/** 
	 *	Find an Object3D in the scene by tag.
	 *	@param {string} tag The tag to search by.
	 *	@returns {Array} Array of objects found with the tag
	 */
	this.findByTag = function(tag){
		var gameObjects = [];
		var stack = [];
		stack.push(this.scene);
		
		
		while (stack.length > 0)
		{
			var currentObject = stack.pop();
			
			if (currentObject.tag == tag){
				gameObjects.push(currentObject);
			}
			stack = stack.concat(currentObject.children);
		}
		return gameObjects;
	};
};
