/** 
 *  Input Class
 *  @class
 *	@classdesc Library to keep track of all user input.  NOTE: You will never need to make a new Input object.  Please refer to the instance at ENGINE.Input.
 *  @param {string} canvas The canvas to check for input on.
 */
var Input = function(canvas){

	/** KeyCode/Mouse Button Enumeration
	* @readonly
 	* @enum {number}
	*/
	this.KEY = 
	{
		/** @ignore */MOUSE1: 1,
		/** @ignore */MOUSE2: 2,
		/** @ignore */MOUSE3: 3,
		/** @ignore */MWHEEL_UP: 4,
		/** @ignore */MWHEEL_DOWN: 5,

		/** @ignore */BACKSPACE: 		8,
		/** @ignore */TAB: 			9,
		/** @ignore */ENTER: 			13,
		/** @ignore */SHIFT: 			16,
		/** @ignore */CTRL: 			17,
		/** @ignore */ALT: 			18,
		/** @ignore */PAUSE: 			19,
		/** @ignore */CAPS: 			20,
		/** @ignore */ESC: 			27,
		/** @ignore */SPACE: 			32,
		/** @ignore */PAGE_UP: 		33,
		/** @ignore */PAGE_DOWN: 		34,
		/** @ignore */END: 			35,
		/** @ignore */HOME: 			36,
		/** @ignore */LEFT_ARROW: 	37,
		/** @ignore */UP_ARROW: 		38,
		/** @ignore */RIGHT_ARROW: 	39,
		/** @ignore */DOWN_ARROW: 	40,
		/** @ignore */INSERT: 		45,
		/** @ignore */DELETE: 		46,
		/** @ignore */_0: 			48,
		/** @ignore */_1:				49,
		/** @ignore */_2: 			50,
		/** @ignore */_3: 			51,
		/** @ignore */_4:			 	52,
		/** @ignore */_5: 			53,
		/** @ignore */_6: 			54,
		/** @ignore */_7: 			55,
		/** @ignore */_8: 			56,
		/** @ignore */_9: 			57,
		/** @ignore */A:				65,
		/** @ignore */B:				66,
		/** @ignore */C:				67,
		/** @ignore */D:				68,
		/** @ignore */E:				69,
		/** @ignore */F:				70,
		/** @ignore */G:				71,
		/** @ignore */H:				72,
		/** @ignore */I:				73,
		/** @ignore */J:				74,
		/** @ignore */K:				75,
		/** @ignore */L:				76,
		/** @ignore */M:				77,
		/** @ignore */N:				78,
		/** @ignore */O:				79,
		/** @ignore */P:				80,
		/** @ignore */Q:				81,
		/** @ignore */R:				82,
		/** @ignore */S:				83,
		/** @ignore */T:				84,
		/** @ignore */U:				85,
		/** @ignore */V:				86,
		/** @ignore */W:				87,
		/** @ignore */X:				88,
		/** @ignore */Y:				89,
		/** @ignore */Z:				90,
		/** @ignore */NUMPAD_0:		96,
		/** @ignore */NUMPAD_1:		97,
		/** @ignore */NUMPAD_2:		98,
		/** @ignore */NUMPAD_3:		99,
		/** @ignore */NUMPAD_4:		100,
		/** @ignore */NUMPAD_5:		101,
		/** @ignore */NUMPAD_6:		102,
		/** @ignore */NUMPAD_7:		103,
		/** @ignore */NUMPAD_8:		104,
		/** @ignore */NUMPAD_9:		105,
		/** @ignore */MULTIPLY:		106,
		/** @ignore */ADD:			107,
		/** @ignore */SUBSTRACT:		109,
		/** @ignore */DECIMAL:		110,
		/** @ignore */DIVIDE:			111,
		/** @ignore */F1:				112,
		/** @ignore */F2:				113,
		/** @ignore */F3:				114,
		/** @ignore */F4:				115,
		/** @ignore */F5:				116,
		/** @ignore */F6:				117,
		/** @ignore */F7:				118,
		/** @ignore */F8:				119,
		/** @ignore */F9:				120,
		/** @ignore */F10:			121,
		/** @ignore */F11:			122,
		/** @ignore */F12:			123,
		/** @ignore */NUM_LOCK:		144,
		/** @ignore */SCROLL_LOCK:	145,
		/** @ignore */SEMI_COLON:		186,
		/** @ignore */EQUALS:			187,
		/** @ignore */COMMA:			188,
		/** @ignore */DASH:			189,
		/** @ignore */PERIOD:			190,
		/** @ignore */FORWARD_SLASH:	191,
		/** @ignore */GRAVE:			192,
		/** @ignore */OPEN_BRACKET:	219,
		/** @ignore */BACK_SLASH:		220,
		/** @ignore */CLOSE_BRACKET:	221,
		/** @ignore */SINGLE_QUOTE:	222,
	};

	// Private Data Structures
	previousKeyState = new Array();
	currentKeyState = new Array();
	for(var i = 0; i < 500; i++)
	{
		previousKeyState[i] = false;
		currentKeyState[i] = false;
	}
	
	this.gamepadSupport = gamepadSupport;
	this.gamepadSupport.previousKeyState = previousKeyState;
	this.gamepadSupport.currentKeyState = currentKeyState;
	
	var mouseCoordinates = new THREE.Vector2(0,0);

	/**
     *	Mouse Coordinates
     *  @readonly
     */
	this.Mouse = {};
	/**
     *	X Coordinate in Pixels
     *  @type number
     */
	this.Mouse.x = function(){return mouseCoordinates.x;};
	/**
     *	Y Coordinate in Pixels
     *  @type number
     */
	this.Mouse.y = function(){return mouseCoordinates.y;};

	/**
	 * Gamepad
	 * Polling code taken from GamepadHandler.js by Jens Arps
	 *
	 * https://github.com/jensarps/decoupled-input/blob/master/src/GamepadHandler.js
	 */
	this.GAMEPAD = {
		BUTTONS : {
			FACE_1: 0, // Face (main) buttons
			FACE_2: 1,
			FACE_3: 2,
			FACE_4: 3,
			LEFT_SHOULDER: 4, // Top shoulder buttons
			RIGHT_SHOULDER: 5,
			LEFT_SHOULDER_BOTTOM: 6, // Bottom shoulder buttons
			RIGHT_SHOULDER_BOTTOM: 7,
			SELECT: 8,
			START: 9,
			LEFT_ANALOGUE_STICK: 10, // Analogue sticks (if depressible)
			RIGHT_ANALOGUE_STICK: 11,
			PAD_TOP: 12, // Directional (discrete) pad
			PAD_BOTTOM: 13,
			PAD_LEFT: 14,
			PAD_RIGHT: 15
		},

		AXES : { 
			LEFT_ANALOG_HOR: 16,
			LEFT_ANALOG_VERT: 17,
			RIGHT_ANALOG_HOR: 18,
			RIGHT_ANALOG_VERT: 19
		},
		
		getBindingKey : function(binding, playerIndex) {
			// console.log("Binding: ", binding);
			return gamepadSupport.KEY_STATE_OFFSET + (playerIndex - 1) * 20 + binding;
		}
	};

	// Hook Functions to Events
	document.onkeydown = function(e){
		currentKeyState[e.keyCode] = true;
		e.preventDefault(); // Prevent default key action (e.g. scrolling with arrow keys)
	};
	document.onkeyup = function(e){
		currentKeyState[e.keyCode] = false;
		e.preventDefault();
	};
	
	window.onblur = function(e){
		for(var i = 0; i < 500; i++)
		{
			currentKeyState[i] = false;
		}
	};
	
	canvas.onmousemove = function(e) {
		// TODO: UV positions?
        var rect = canvas.getBoundingClientRect();
		mouseCoordinates.x = e.clientX - rect.left;
		mouseCoordinates.y = e.clientY - rect.top;
    };
	canvas.onmousedown = function(e) {

		currentKeyState[e.which]=true;
		e.preventDefault();
	};
	
	canvas.onmouseup = function(e) {
		currentKeyState[e.which]=false;
		e.preventDefault();
	};
	canvas.onmousewheel = function(e){
		if(e.wheelDelta >= 120)
		{
			currentKeyState[ENGINE.Input.KEY.MWHEEL_UP] = true;
		}
		else if (e.wheelDelta <= -120)
		{
			currentKeyState[ENGINE.Input.KEY.MWHEEL_DOWN] = true;
		}
		e.preventDefault();
	};

	var binds = {};

	this.init = function() {
		this.gamepadSupport.init();
	}

	// TODO
	this.loadConfiguration = function(configuration){
		
	};

	this.updatePreviousKeyState = function(){
		// Copy By Value because they are bools
		previousKeyState = currentKeyState.slice(0);
	};
	
	/**
     *	Bind a key to some action.
     *  @param {string} action The name of the action you wish to bind a key to.
	 *  @param {number} key The code of the key you wish to bind to the action.
     */
	this.addBind = function(action, key)
	{
		// console.log(key);
		// TODO: check if key valid
		if(binds[action] !== undefined)
		{
			if(!binds[action].indexOf(key) == -1)
			{
				binds[action].push(key);
			}
		}
		else
		{
			binds[action] = [key];
		}
	};
	
	/** 
	 *  Remove a specific binding by action name
	 *  @param {string} action The action you wish to remove
	 */
	this.removeBind = function(action){
		if(binds.action !== undefined)
		{
			delete binds.action;	
		}
	};

	/** 
	 *  Removes all current input bindings.
	 */
	this.removeAllBinds = function(){
		binds = {};
	};
	
	/** 
	 *  Determine whether a key is currently down.
	 *  @param {string} action The action to check.
	 *	@returns {Boolean}
	 */
	this.state = function(action){
		var keys = binds[action];
		if (keys !== undefined) {
			for(var i = 0; i < keys.length; i++)
			{
				var key = keys[i];
				return currentKeyState[key];
			}
		}
		return false;
	};

	/** 
	 *  Determine whether a key was just pressed this frame.
	 *  @param {string} action The action to check.
	 *	@returns {Boolean}
	 */
	this.pressed = function(action){
		var keys = binds[action];
		if (keys !== undefined) {
			for(var i = 0; i < keys.length; i++)
			{
				var key = keys[i];
				if(currentKeyState[key] && !previousKeyState[key]){
					return true;
				}
			}
		}
		return false;
	};

	/** 
	 *  Determine whether a key was just released this frame.
	 *  @param {string} action The action to check.
	 *	@returns {Boolean}
	 */
	this.released = function(action){
		var keys = binds[action];
		if (keys !== undefined) {
			for(var i = 0; i < keys.length; i++)
			{
				var key = keys[i];
				if(!currentKeyState[key] && previousKeyState[key]){
					return true;
				}
			}
		}
		return false;
	};
};

