/**
 * Object for a High Score Table.
 * @param maxScores (int) The number for maximum scores to save/display. Default: 10
 * @param saveOnAdd (bool) Set to true if the table should save after every add operation. Default: false
 */
function HighScoreTable(maxScores, saveOnAdd)
{
	this.size = maxScores || 10;
	this.autoSave = saveOnAdd || false;
	
	this.defaultName = "Andy";
	this.maxNameSize = 20;
	this.maxScoreSize = 10;
	
	this.scores = [];
	this.scoresSortFunction = function(a, b) { return b.score - a.score; }
	
	for(var i = 0; i < this.size; i++)
	{
		this.scores.push({name:this.defaultName, score:0});
	}
}

/**
 * Adds a score to the table
 * @param name (string) The name of the player
 * @param score (int) The score of the player
 */
HighScoreTable.prototype.addScore = function(name, score)
{
	name = name || "Andy";
	score = score || 0;
	
	this.scores.push({name:name, score:score});
	this.scores.sort(this.scoresSortFunction);
	this.scores.length = this.size;
	
	if(this.autoSave)
	{
		this.saveScores();
	}
}

/**
 * Loads scores from local storage
 */
HighScoreTable.prototype.loadScores = function()
{
	this.size = parseInt(localStorage.size);
	
	//override autosave, load it after adding scores.
	this.autoSave = false;
	
	for(var i = 0; i < this.size; i++)
	{
		this.addScore(localStorage['name'+i], parseInt(localStorage['score'+i]));
	}
	
	this.autoSave = localStorage.autoSave !== 'false';
}

/**
 * Saves scores to local storage
 * WARNING: multiple objects will result in saving to the same location.
 */
HighScoreTable.prototype.saveScores = function()
{
	localStorage.size = this.size;
	localStorage.autoSave = this.autoSave;
	
	for(var i = 0; i < this.size; i++)
	{
		localStorage['name'+i] = this.scores[i].name;
		localStorage['score'+i] = this.scores[i].score;
	}
}

/**
 * Gets a string of scores to display
 * @param displayType (int) The type to display. Options: SCORE_DISPLAY_TEXT (end in \n), SCORE_DISPLAY_HTML (<br/>), SCORE_DISPLAY_UL, SCORE_DISPLAY_OL)
 * @returns a string of the scores to display
 */
HighScoreTable.prototype.displayScores = function(displayType)
{
	displayType = displayType || SCORE_DISPLAY_HTML;
	
	var output = "";
	
	for(var i = 0; i < this.size; i++)
	{
		var name = this.scores[i].name;
		var score = this.scores[i].score + ""; //stringify it
		
		var dots = (this.maxNameSize - name.length) + (this.maxScoreSize - score.length);
		
		var string = name + Array(dots+1).join('.') + score;
		
		switch(displayType)
		{
			case SCORE_DISPLAY_HTML:
				string += "<br/>";
				break;
			case SCORE_DISPLAY_TEXT:
				string += "\n";
				break;
			case SCORE_DISPLAY_OL:
			case SCORE_DISPLAY_UL:
				string = "<li>" + string + "</li>";
				break;
		}
		output += string;
	}
	
	if(displayType == SCORE_DISPLAY_OL)
	{
		output = "<ol>" + output + "</ol>";
	}
	else if (displayType == SCORE_DISPLAY_UL)
	{
		output = "<ul>" + output + "</ul>";
	}
	
	return output;
}

var SCORE_DISPLAY_HTML = 0;
var SCORE_DISPLAY_TEXT = 1;
var SCORE_DISPLAY_OL = 2;
var SCORE_DISPLAY_UL = 3;

var highScoreTable = new HighScoreTable();